/* Fo
 * fo-repeatable-page-master-reference.c: Repeatable Page Master Reference formatting object
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-repeatable-page-master-reference.c,v 1.2 2003/03/19 21:12:08 tonygraham Exp $
 *
 * See Copyright for the status of this software.
 */

#include "libfo/fo-utils.h"
#include "fo/fo-fo.h"
#include "fo/fo-fo-private.h"
#include "fo-repeatable-page-master-reference.h"
#include "fo-repeatable-page-master-reference-private.h"

enum {
  PROP_0,
  PROP_MASTER_REFERENCE
};

static void fo_repeatable_page_master_reference_init        (FoRepeatablePageMasterReference      *fo_repeatable_page_master_reference);
static void fo_repeatable_page_master_reference_class_init  (FoRepeatablePageMasterReferenceClass *klass);
static void fo_repeatable_page_master_reference_set_property (GObject         *object,
                                  guint            prop_id,
                                  const GValue    *value,
                                  GParamSpec      *pspec);
static void fo_repeatable_page_master_reference_get_property   (GObject         *object,
                                       guint            prop_id,
                                       GValue          *value,
                                       GParamSpec      *pspec);
static void fo_repeatable_page_master_reference_finalize    (GObject           *object);
static void fo_repeatable_page_master_reference_debug_dump_properties (FoFo *fo, gint depth);

static gpointer parent_class;

GType
fo_repeatable_page_master_reference_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoRepeatablePageMasterReferenceClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) fo_repeatable_page_master_reference_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoRepeatablePageMasterReference),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_repeatable_page_master_reference_init,
      };
      
      object_type = g_type_register_static (FO_TYPE_FO,
                                            "FoRepeatablePageMasterReference",
                                            &object_info, 0);
    }
  
  return object_type;
}

static void
fo_repeatable_page_master_reference_init (FoRepeatablePageMasterReference *fo_repeatable_page_master_reference)
{
}

static void
fo_repeatable_page_master_reference_class_init (FoRepeatablePageMasterReferenceClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);
  
  object_class->finalize = fo_repeatable_page_master_reference_finalize;

  object_class->set_property = fo_repeatable_page_master_reference_set_property;
  object_class->get_property = fo_repeatable_page_master_reference_get_property;

  FO_FO_CLASS (klass)->debug_dump_properties = fo_repeatable_page_master_reference_debug_dump_properties;
  FO_FO_CLASS (klass)->update_from_context = fo_repeatable_page_master_reference_update_from_context;

  g_object_class_install_property
    (object_class,
     PROP_MASTER_REFERENCE,
     g_param_spec_object ("master-reference",
			  _("Master Reference"),
			  _("Master Reference property"),
			  FO_TYPE_PROPERTY,
			  G_PARAM_READABLE));
}

static void
fo_repeatable_page_master_reference_finalize (GObject *object)
{
  FoRepeatablePageMasterReference *fo_repeatable_page_master_reference;

  fo_repeatable_page_master_reference = FO_REPEATABLE_PAGE_MASTER_REFERENCE (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


static void
fo_repeatable_page_master_reference_set_property (GObject         *object,
                         guint            prop_id,
                         const GValue    *value,
                         GParamSpec      *pspec)
{
  FoRepeatablePageMasterReference *fo_repeatable_page_master_reference;

  fo_repeatable_page_master_reference = FO_REPEATABLE_PAGE_MASTER_REFERENCE (object);

  switch (prop_id)
    {
    case PROP_MASTER_REFERENCE:
      fo_repeatable_page_master_reference_set_master_reference (fo_repeatable_page_master_reference, FO_PROPERTY_MASTER_REFERENCE (g_value_get_object (value)));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
fo_repeatable_page_master_reference_get_property (GObject         *object,
                         guint            prop_id,
                         GValue          *value,
                         GParamSpec      *pspec)
{
  FoRepeatablePageMasterReference *fo_repeatable_page_master_reference;

  fo_repeatable_page_master_reference = FO_REPEATABLE_PAGE_MASTER_REFERENCE (object);

  switch (prop_id)
    {
    case PROP_MASTER_REFERENCE:
      g_value_set_object (value, G_OBJECT (fo_repeatable_page_master_reference_get_master_reference (fo_repeatable_page_master_reference)));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

/**
 * fo_repeatable_page_master_reference_new:
 * 
 * Creates a new #FoRepeatablePageMasterReference initialized to default value.
 * 
 * Return value: the new #FoRepeatablePageMasterReference
 **/
FoFo*
fo_repeatable_page_master_reference_new (void)
{
  return FO_FO (g_object_new (fo_repeatable_page_master_reference_get_type (), NULL));
}

void
fo_repeatable_page_master_reference_update_from_context (FoFo *fo, FoContext *context)
{
  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_REPEATABLE_PAGE_MASTER_REFERENCE (fo));

  fo_repeatable_page_master_reference_set_master_reference (FO_REPEATABLE_PAGE_MASTER_REFERENCE (fo),
			  fo_context_get_master_reference (context));
}

void
fo_repeatable_page_master_reference_debug_dump_properties (FoFo *fo, gint depth)
{
  FoRepeatablePageMasterReference *fo_repeatable_page_master_reference;

  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_REPEATABLE_PAGE_MASTER_REFERENCE (fo));

  fo_repeatable_page_master_reference = FO_REPEATABLE_PAGE_MASTER_REFERENCE (fo);

  fo_object_debug_dump (fo_repeatable_page_master_reference->master_reference, depth);

  FO_FO_CLASS (parent_class)->debug_dump_properties (fo, depth + 1);
}

/**
 * fo_repeatable_page_master_reference_set_master_reference:
 * @fo_repeatable_page_master_reference: The @FoRepeatablePageMasterReference object
 * @new_master_reference: The new "master-reference" property value
 * 
 * Sets the #master-reference property of @repeatable_page_master_reference to @new_master_reference
 **/
void
fo_repeatable_page_master_reference_set_master_reference (FoRepeatablePageMasterReference *fo_repeatable_page_master_reference,
		         FoPropertyMasterReference *new_master_reference)
{
  g_return_if_fail (fo_repeatable_page_master_reference != NULL);
  g_return_if_fail (FO_IS_REPEATABLE_PAGE_MASTER_REFERENCE (fo_repeatable_page_master_reference));

  if (new_master_reference)
    g_object_ref (G_OBJECT (new_master_reference));
  if (fo_repeatable_page_master_reference->master_reference)
    g_object_unref (G_OBJECT (fo_repeatable_page_master_reference->master_reference));
  fo_repeatable_page_master_reference->master_reference = FO_PROPERTY_MASTER_REFERENCE (new_master_reference);
  /*g_object_notify (G_OBJECT (fo_repeatable_page_master_reference), "master-reference");*/
}

/**
 * fo_repeatable_page_master_reference_get_master_reference:
 * @fo_repeatable_page_master_reference: The @FoRepeatablePageMasterReference object
 *
 * Gets the #master-reference property of @repeatable_page_master_reference
 *
 * Return value: The "master-reference" property value
**/
FoPropertyMasterReference*
fo_repeatable_page_master_reference_get_master_reference (FoRepeatablePageMasterReference *fo_repeatable_page_master_reference)
{
  g_return_val_if_fail (fo_repeatable_page_master_reference != NULL, NULL);
  g_return_val_if_fail (FO_IS_REPEATABLE_PAGE_MASTER_REFERENCE (fo_repeatable_page_master_reference), NULL);

  return fo_repeatable_page_master_reference->master_reference;
}
