/* Fo
 * fo-property-font-weight.c: 'font-weight' property
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-property-font-weight.c,v 1.3 2003/03/21 14:41:31 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include <string.h>
#include "libfo/fo-utils.h"
#include "fo-context.h"
#include "datatype/fo-datatype.h"
#include "property/fo-property-private.h"
#include "property/fo-property-text-property.h"
#include "property/fo-property-font-size.h"
#include "property/fo-property-font-weight.h"

#ifndef FO_FONT_WEIGHT_NORMAL
#define FO_FONT_WEIGHT_NORMAL	400
#endif	/* !FO_FONT_WEIGHT_NORMAL */

#ifndef FO_FONT_WEIGHT_BOLD
#define FO_FONT_WEIGHT_BOLD	700
#endif	/* !FO_FONT_WEIGHT_BOLD */

#ifndef FO_FONT_WEIGHT_STEP
#define FO_FONT_WEIGHT_STEP	100
#endif	/* !FO_FONT_WEIGHT_STEP */

#ifndef FO_FONT_WEIGHT_MAX
#define FO_FONT_WEIGHT_MAX	900
#endif	/* !FO_FONT_WEIGHT_MAX */

#ifndef FO_FONT_WEIGHT_MIN
#define FO_FONT_WEIGHT_MIN	100
#endif	/* !FO_FONT_WEIGHT_MIN */

/* Inherited: TRUE */
/* Shorthand: FALSE */
/* font-weight */
/* normal | bold | bolder | lighter | 100 | 200 | 300 | 400 | 500 | 600 | 700 | 800 | 900 | inherit */
GType
fo_enum_font_weight_get_type (void)
{
  static GType etype = 0;
  if (etype == 0)
    {
      static const GEnumValue values[] = {
        { FO_ENUM_FONT_WEIGHT_NORMAL, "FO_ENUM_FONT_WEIGHT_NORMAL", "normal" },
        { FO_ENUM_FONT_WEIGHT_BOLD, "FO_ENUM_FONT_WEIGHT_BOLD", "bold" },
        { FO_ENUM_FONT_WEIGHT_BOLDER, "FO_ENUM_FONT_WEIGHT_BOLDER", "bolder" },
        { FO_ENUM_FONT_WEIGHT_LIGHTER, "FO_ENUM_FONT_WEIGHT_LIGHTER", "lighter" },
        { FO_ENUM_FONT_WEIGHT_100, "FO_ENUM_FONT_WEIGHT_100", "100" },
        { FO_ENUM_FONT_WEIGHT_200, "FO_ENUM_FONT_WEIGHT_200", "200" },
        { FO_ENUM_FONT_WEIGHT_300, "FO_ENUM_FONT_WEIGHT_300", "300" },
        { FO_ENUM_FONT_WEIGHT_400, "FO_ENUM_FONT_WEIGHT_400", "400" },
        { FO_ENUM_FONT_WEIGHT_500, "FO_ENUM_FONT_WEIGHT_500", "500" },
        { FO_ENUM_FONT_WEIGHT_600, "FO_ENUM_FONT_WEIGHT_600", "600" },
        { FO_ENUM_FONT_WEIGHT_700, "FO_ENUM_FONT_WEIGHT_700", "700" },
        { FO_ENUM_FONT_WEIGHT_800, "FO_ENUM_FONT_WEIGHT_800", "800" },
        { FO_ENUM_FONT_WEIGHT_900, "FO_ENUM_FONT_WEIGHT_900", "900" },
        { 0, NULL, NULL }
      };
      etype = g_enum_register_static ("FoEnumFontWeight", values);
    }
  return etype;
}

struct _FoPropertyFontWeight
{
  FoProperty parent_instance;
};

struct _FoPropertyFontWeightClass
{
  FoPropertyClass parent_class;
};

static void fo_property_font_weight_init         (FoPropertyFontWeight      *property_font_weight);
static void fo_property_font_weight_base_init   (FoPropertyFontWeightClass *klass);
static void fo_property_font_weight_base_finalize   (FoPropertyFontWeightClass *klass);
static void fo_property_font_weight_class_init   (FoPropertyFontWeightClass *klass);
static void fo_property_font_weight_finalize     (GObject       *object);
static void fo_property_font_weight_text_property_init (FoPropertyTextPropertyIface *iface);

static FoDatatype* fo_property_font_weight_resolve_enum (const gchar *token,
                                                         FoContext   *context,
                                                         GError     **error);
static FoDatatype* fo_property_font_weight_validate (FoDatatype *datatype,
                                                     FoContext  *context,
                                                     GError    **error);
static PangoAttribute * fo_property_font_weight_new_attr (FoProperty *property);

static const gchar class_name[] = "font-weight";
static gpointer parent_class;
static FoDatatype *enum_100;
static FoDatatype *enum_200;
static FoDatatype *enum_300;
static FoDatatype *enum_400;
static FoDatatype *enum_500;
static FoDatatype *enum_600;
static FoDatatype *enum_700;
static FoDatatype *enum_800;
static FoDatatype *enum_900;

/**
 * fo_property_font_weight_get_type:
 * 
 * Register the #FoPropertyFontWeight type if not already registered and
 * return its #GType value.
 * 
 * Return value: #GType of #FoPropertyFontWeight.
 **/
GType
fo_property_font_weight_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoPropertyFontWeightClass),
        (GBaseInitFunc) fo_property_font_weight_base_init,
        (GBaseFinalizeFunc) fo_property_font_weight_base_finalize,
        (GClassInitFunc) fo_property_font_weight_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoPropertyFontWeight),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_property_font_weight_init,
	NULL		/* value_table */
      };

      static const GInterfaceInfo fo_property_text_property_info =
      {
	(GInterfaceInitFunc) fo_property_font_weight_text_property_init, /* interface_init */
        NULL,
        NULL
      };

      object_type = g_type_register_static (FO_TYPE_PROPERTY,
                                            class_name,
                                            &object_info, 0);
      g_type_add_interface_static (object_type,
                                   FO_TYPE_PROPERTY_TEXT_PROPERTY,
                                   &fo_property_text_property_info);
    }

  return object_type;
}

/**
 * fo_property_font_weight_init:
 * @font_weight: #FoPropertyFontWeight object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoPropertyFontWeight.
 **/
void
fo_property_font_weight_init (FoPropertyFontWeight *font_weight)
{
  FO_PROPERTY (font_weight)->value =
    g_object_ref (enum_400);
}

/**
 * fo_property_font_weight_base_init:
 * @klass: #FoPropertyFontWeightClass object to initialise.
 * 
 * Implements #GBaseInitFunc for #FoPropertyFontWeightClass.
 **/
void
fo_property_font_weight_base_init (FoPropertyFontWeightClass *klass)
{
  enum_100 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  100,
		  NULL);
  enum_200 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  200,
		  NULL);
  enum_300 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  300,
		  NULL);
  enum_400 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  400,
		  NULL);
  enum_500 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  500,
		  NULL);
  enum_600 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  600,
		  NULL);
  enum_700 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  700,
		  NULL);
  enum_800 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  800,
		  NULL);
  enum_900 =
    g_object_new (FO_TYPE_INTEGER,
		  "value",
		  900,
		  NULL);
}

/**
 * fo_property_font_weight_base_finalize:
 * @klass: #FoPropertyFontWeightClass object to finalise.
 * 
 * Implements #GBaseFinalizeFunc for #FoPropertyFontWeightClass.
 **/
void
fo_property_font_weight_base_finalize (FoPropertyFontWeightClass *klass)
{
  g_object_unref (enum_100);
  g_object_unref (enum_200);
  g_object_unref (enum_300);
  g_object_unref (enum_400);
  g_object_unref (enum_500);
  g_object_unref (enum_600);
  g_object_unref (enum_700);
  g_object_unref (enum_800);
  g_object_unref (enum_900);
}

/**
 * fo_property_font_weight_class_init:
 * @klass: #FoPropertyFontWeightClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoPropertyFontWeightClass.
 **/
void
fo_property_font_weight_class_init (FoPropertyFontWeightClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  FoPropertyClass *property_class = FO_PROPERTY_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = fo_property_font_weight_finalize;

  property_class->is_inherited = TRUE;
  property_class->is_shorthand = FALSE;
  property_class->resolve_enum =
    fo_property_font_weight_resolve_enum;
  property_class->validate =
    fo_property_font_weight_validate;
  property_class->get_initial =
    fo_property_font_weight_get_initial;
}

/**
 * fo_property_font_weight_finalize:
 * @object: #FoPropertyFontWeight object to finalize.
 * 
 * Implements #GObjectFinalizeFunc for #FoPropertyFontWeight.
 **/
void
fo_property_font_weight_finalize (GObject *object)
{
  FoPropertyFontWeight *font_weight;

  font_weight = FO_PROPERTY_FONT_WEIGHT (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * fo_property_font_weight_text_property_init:
 * @iface: #FoPropertyTextPropertyIFace structure for this class
 * 
 * Initialize #FoPropertyTextPropertyIface interface for this class
 **/
void
fo_property_font_weight_text_property_init (FoPropertyTextPropertyIface *iface)
{
  iface->new_attr = fo_property_font_weight_new_attr;
}

/**
 * fo_property_font_weight_new:
 * 
 * Creates a new #FoPropertyFontWeight initialized to default value.
 * 
 * Return value: the new #FoPropertyFontWeight.
 **/
FoProperty*
fo_property_font_weight_new (void)
{
  FoProperty* font_weight;

  font_weight =
    FO_PROPERTY (g_object_new (fo_property_font_weight_get_type (),
                               NULL));

  return font_weight;
}

/**
 * fo_property_font_weight_resolve_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_font_weight_resolve_enum (const gchar *token,
                                      FoContext   *context,
                                      GError     **error)
{
  gint old_font_weight_value;

  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "normal") == 0)
    {
      return g_object_ref (enum_400);
    }
  else if (strcmp (token, "bold") == 0)
    {
      return g_object_ref (enum_700);
    }
  else if (strcmp (token, "bolder") == 0)
    {
      /* FIXME: Redo with fo_datatype_max(), etc. */
      old_font_weight_value =
	fo_integer_get_value (fo_property_get_value (fo_context_get_font_weight (context)));

      return g_object_new (FO_TYPE_INTEGER,
			   "value",
			   MIN (old_font_weight_value + FO_FONT_WEIGHT_STEP,
				FO_FONT_WEIGHT_MAX),
			   NULL);
    }
  else if (strcmp (token, "lighter") == 0)
    {
      old_font_weight_value =
	fo_integer_get_value (fo_property_get_value (fo_context_get_font_weight (context)));

      return g_object_new (FO_TYPE_INTEGER,
			   "value",
			   MAX (old_font_weight_value - FO_FONT_WEIGHT_STEP,
				FO_FONT_WEIGHT_MIN),
			   NULL);
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   class_name,
		   token);
      return NULL;
    }
}

/**
 * fo_property_font_weight_validate:
 * @datatype: #FoDatatype to be validated against allowed datatypes and
 *            values for current property.
 * @context:  #FoContext object from which to possibly inherit values.
 * @error:    Information about any error that has occurred.
 * 
 * Validates @datatype against allowed values.  Returns @datatype, a
 * replacement datatype value, or NULL if validation failed.
 * 
 * Return value: Valid datatype value or NULL.
 **/
FoDatatype*
fo_property_font_weight_validate (FoDatatype *datatype,
                                  FoContext  *context,
                                  GError    **error)
{
  FoDatatype *new_datatype;
  GError     *tmp_error = NULL;
  gchar      *token;
  gint        int_value;
  gfloat      float_value;

  g_return_val_if_fail (datatype != NULL, NULL);
  g_return_val_if_fail (FO_IS_DATATYPE (datatype), NULL);
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (FO_IS_ENUM (datatype))
    {
      if ((G_ENUM_CLASS_TYPE (fo_enum_get_enum_class (datatype))
           == FO_TYPE_ENUM_FONT_WEIGHT) &&
	  (g_enum_get_value (fo_enum_get_enum_class (datatype),
			     fo_enum_get_value (datatype)) != NULL))
	{
	  return datatype;
	}
      else
	{
	  gchar *datatype_sprintf = fo_object_sprintf (datatype);

	  g_set_error (error,
		       FO_FO_ERROR,
		       FO_FO_ERROR_DATATYPE,
		       _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		       class_name,
		       datatype_sprintf,
		       g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

	  g_object_unref (datatype);

	  g_free (datatype_sprintf);

	  return NULL;
	}
    }
  else if (FO_IS_STRING (datatype))
    {
      token = fo_string_get_value (datatype);

      new_datatype =
        fo_property_font_weight_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else if (FO_IS_NAME (datatype))
    {
      token = fo_name_get_value (datatype);

      new_datatype =
        fo_property_font_weight_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else if (FO_IS_NUMBER (datatype))
    {
      float_value = fo_number_get_value (datatype);

      if ((float_value == 100) || (float_value == 200) ||
	  (float_value == 300) || (float_value == 400) ||
	  (float_value == 500) || (float_value == 600) ||
	  (float_value == 700) || (float_value == 800) ||
	  (float_value == 900))
	{
	  new_datatype = g_object_new (FO_TYPE_INTEGER,
				       "value",
				       (int) float_value,
				       NULL);
	  g_object_unref (datatype);

	  return new_datatype;
	}
      else
	{
	  g_warning ("Incorrect font-weight: %g", float_value);

	  int_value = (int) float_value / 100;

	  int_value = CLAMP (int_value, 1, 9) * 100;

	  new_datatype = g_object_new (FO_TYPE_INTEGER,
				       "value",
				       int_value,
				       NULL);

	  g_object_unref (datatype);

	  return new_datatype;
	}
    }
  else if (FO_IS_INTEGER (datatype))
    {
      int_value = fo_integer_get_value (datatype);

      if ((int_value == 100) || (int_value == 200) ||
	  (int_value == 300) || (int_value == 400) ||
	  (int_value == 500) || (int_value == 600) ||
	  (int_value == 700) || (int_value == 800) ||
	  (int_value == 900))
	{
	  return datatype;
	}
      else
	{
	  g_warning ("Incorrect font-weight: %d", int_value);

	  int_value = CLAMP (int_value / 100, 1, 9) * 100;

	  new_datatype = g_object_new (FO_TYPE_INTEGER,
				       "value",
				       int_value,
				       NULL);

	  g_object_unref (datatype);

	  return new_datatype;
	}
    }
  else
    {
      gchar *datatype_sprintf = fo_object_sprintf (datatype);

      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_DATATYPE,
		   _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		   class_name,
		   datatype_sprintf,
		   g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

      g_object_unref (datatype);

      g_free (datatype_sprintf);

      return NULL;
    }
}

/**
 * fo_property_font_weight_get_initial:
 * 
 * Get an instance of the property with the correct initial value.
 * 
 * Return value: An instance of the property.
 **/
FoProperty*
fo_property_font_weight_get_initial (void)
{
  static FoProperty *font_weight = NULL;

  if (font_weight == NULL)
    {
      font_weight = fo_property_font_weight_new ();
    }

  return font_weight;
}

/**
 * fo_property_font_weight_new_attr:
 * @property: #FoProperty from which to create a #PangoAttribute.
 * 
 * Makes a new #PangoAttribute representing the current property.
 * 
 * Return value: New #PangoAttribute.
 **/
PangoAttribute*
fo_property_font_weight_new_attr (FoProperty *property)
{
  gint pango_font_weight;

  g_return_val_if_fail (FO_IS_PROPERTY_FONT_WEIGHT (property), NULL);

  pango_font_weight =
    fo_integer_get_value (property->value);

  return pango_attr_weight_new (pango_font_weight);
}
