/* Fo
 * fo-property-width.c: 'width' property
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-property-width.c,v 1.3 2003/07/02 10:39:38 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include <string.h>
#include "libfo/fo-utils.h"
#include "fo-context.h"
#include "datatype/fo-datatype.h"
#include "property/fo-property-private.h"
#include "property/fo-property-font-size.h"
#include "property/fo-property-width.h"

/* Inherited: FALSE */
/* Shorthand: FALSE */
/* width */
/* <length> | <percentage> | auto | inherit */
GType
fo_enum_width_get_type (void)
{
  static GType etype = 0;
  if (etype == 0)
    {
      static const GEnumValue values[] = {
        { FO_ENUM_WIDTH_AUTO, "FO_ENUM_WIDTH_AUTO", "auto" },
        { 0, NULL, NULL }
      };
      etype = g_enum_register_static ("FoEnumWidth", values);
    }
  return etype;
}

struct _FoPropertyWidth
{
  FoProperty parent_instance;
};

struct _FoPropertyWidthClass
{
  FoPropertyClass parent_class;
};

static void fo_property_width_init         (FoPropertyWidth      *property_width);
static void fo_property_width_base_init   (FoPropertyWidthClass *klass);
static void fo_property_width_base_finalize   (FoPropertyWidthClass *klass);
static void fo_property_width_class_init   (FoPropertyWidthClass *klass);
static void fo_property_width_finalize     (GObject       *object);

static FoDatatype* fo_property_width_resolve_enum (const gchar *token,
                                                   FoContext   *context,
                                                   GError     **error);
static FoDatatype* fo_property_width_validate (FoDatatype *datatype,
                                               FoContext  *context,
                                               GError    **error);

static const gchar class_name[] = "width";
static gpointer parent_class;
static FoDatatype *enum_auto;

/**
 * fo_property_width_get_type:
 * 
 * Register the #FoPropertyWidth type if not already registered and
 * return its #GType value.
 * 
 * Return value: #GType of #FoPropertyWidth.
 **/
GType
fo_property_width_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoPropertyWidthClass),
        (GBaseInitFunc) fo_property_width_base_init,
        (GBaseFinalizeFunc) fo_property_width_base_finalize,
        (GClassInitFunc) fo_property_width_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoPropertyWidth),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_property_width_init,
	NULL		/* value_table */
      };

      object_type = g_type_register_static (FO_TYPE_PROPERTY,
                                            class_name,
                                            &object_info, 0);
    }

  return object_type;
}

/**
 * fo_property_width_init:
 * @width: #FoPropertyWidth object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoPropertyWidth.
 **/
void
fo_property_width_init (FoPropertyWidth *width)
{
  FO_PROPERTY (width)->value =
    g_object_ref (enum_auto);
}

/**
 * fo_property_width_base_init:
 * @klass: #FoPropertyWidthClass object to initialise.
 * 
 * Implements #GBaseInitFunc for #FoPropertyWidthClass.
 **/
void
fo_property_width_base_init (FoPropertyWidthClass *klass)
{
  enum_auto =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_WIDTH),
                  "value",
                  FO_ENUM_WIDTH_AUTO,
                  NULL);
}

/**
 * fo_property_width_base_finalize:
 * @klass: #FoPropertyWidthClass object to finalise.
 * 
 * Implements #GBaseFinalizeFunc for #FoPropertyWidthClass.
 **/
void
fo_property_width_base_finalize (FoPropertyWidthClass *klass)
{
  g_object_unref (enum_auto);
}

/**
 * fo_property_width_class_init:
 * @klass: #FoPropertyWidthClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoPropertyWidthClass.
 **/
void
fo_property_width_class_init (FoPropertyWidthClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  FoPropertyClass *property_class = FO_PROPERTY_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = fo_property_width_finalize;

  property_class->is_inherited = FALSE;
  property_class->is_shorthand = FALSE;
  property_class->resolve_enum =
    fo_property_width_resolve_enum;
  property_class->validate =
    fo_property_width_validate;
  property_class->get_initial =
    fo_property_width_get_initial;
}

/**
 * fo_property_width_finalize:
 * @object: #FoPropertyWidth object to finalize.
 * 
 * Implements #GObjectFinalizeFunc for #FoPropertyWidth.
 **/
void
fo_property_width_finalize (GObject *object)
{
  FoPropertyWidth *width;

  width = FO_PROPERTY_WIDTH (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


/**
 * fo_property_width_new:
 * 
 * Creates a new #FoPropertyWidth initialized to default value.
 * 
 * Return value: the new #FoPropertyWidth.
 **/
FoProperty*
fo_property_width_new (void)
{
  FoProperty* width;

  width =
    FO_PROPERTY (g_object_new (fo_property_width_get_type (),
                               NULL));

  return width;
}

/**
 * fo_property_width_resolve_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_width_resolve_enum (const gchar *token,
                                FoContext   *context,
                                GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "auto") == 0)
    {
      return g_object_ref (enum_auto);
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   class_name,
		   token);
      return NULL;
    }
}

/**
 * fo_property_width_validate:
 * @datatype: #FoDatatype to be validated against allowed datatypes and
 *            values for current property.
 * @context:  #FoContext object from which to possibly inherit values.
 * @error:    Information about any error that has occurred.
 * 
 * Validates @datatype against allowed values.  Returns @datatype, a
 * replacement datatype value, or NULL if validation failed.
 * 
 * Return value: Valid datatype value or NULL.
 **/
FoDatatype*
fo_property_width_validate (FoDatatype *datatype,
                            FoContext  *context,
                            GError    **error)
{
  FoDatatype *new_datatype;
  GError     *tmp_error = NULL;
  gchar      *token;

  g_return_val_if_fail (datatype != NULL, NULL);
  g_return_val_if_fail (FO_IS_DATATYPE (datatype), NULL);
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (FO_IS_LENGTH (datatype))
    {
      return datatype;
    }
  else if (FO_IS_PERCENTAGE (datatype))
    {
      return datatype;
    }
  else if (FO_IS_ENUM (datatype))
    {
      if ((G_ENUM_CLASS_TYPE (fo_enum_get_enum_class (datatype))
           == FO_TYPE_ENUM_WIDTH) &&
	  (g_enum_get_value (fo_enum_get_enum_class (datatype),
			     fo_enum_get_value (datatype)) != NULL))
	{
	  return datatype;
	}
      else
	{
	  gchar *datatype_sprintf = fo_object_sprintf (datatype);

	  g_set_error (error,
		       FO_FO_ERROR,
		       FO_FO_ERROR_DATATYPE,
		       _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		       class_name,
		       datatype_sprintf,
		       g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

	  g_object_unref (datatype);

	  g_free (datatype_sprintf);

	  return NULL;
	}
    }
  else if (FO_IS_STRING (datatype))
    {
      token = fo_string_get_value (datatype);

      new_datatype =
        fo_property_width_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else if (FO_IS_NAME (datatype))
    {
      token = fo_name_get_value (datatype);

      new_datatype =
        fo_property_width_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else
    {
      gchar *datatype_sprintf = fo_object_sprintf (datatype);

      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_DATATYPE,
		   _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		   class_name,
		   datatype_sprintf,
		   g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

      g_object_unref (datatype);

      g_free (datatype_sprintf);

      return NULL;
    }
}

/**
 * fo_property_width_get_initial:
 * 
 * Get an instance of the property with the correct initial value.
 * 
 * Return value: An instance of the property.
 **/
FoProperty*
fo_property_width_get_initial (void)
{
  static FoProperty *width = NULL;

  if (width == NULL)
    {
      width = fo_property_width_new ();
    }

  return width;
}
