/* Fo
 * fo-property-wrap-option.c: 'wrap-option' property
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-property-wrap-option.c,v 1.3 2003/07/02 10:39:38 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include <string.h>
#include "libfo/fo-utils.h"
#include "fo-context.h"
#include "datatype/fo-datatype.h"
#include "property/fo-property-private.h"
#include "property/fo-property-font-size.h"
#include "property/fo-property-wrap-option.h"

/* Inherited: TRUE */
/* Shorthand: FALSE */
/* wrap-option */
/* no-wrap | wrap | inherit */
GType
fo_enum_wrap_option_get_type (void)
{
  static GType etype = 0;
  if (etype == 0)
    {
      static const GEnumValue values[] = {
        { FO_ENUM_WRAP_OPTION_NO_WRAP, "FO_ENUM_WRAP_OPTION_NO_WRAP", "no-wrap" },
        { FO_ENUM_WRAP_OPTION_WRAP, "FO_ENUM_WRAP_OPTION_WRAP", "wrap" },
        { 0, NULL, NULL }
      };
      etype = g_enum_register_static ("FoEnumWrapOption", values);
    }
  return etype;
}

struct _FoPropertyWrapOption
{
  FoProperty parent_instance;
};

struct _FoPropertyWrapOptionClass
{
  FoPropertyClass parent_class;
};

static void fo_property_wrap_option_init         (FoPropertyWrapOption      *property_wrap_option);
static void fo_property_wrap_option_base_init   (FoPropertyWrapOptionClass *klass);
static void fo_property_wrap_option_base_finalize   (FoPropertyWrapOptionClass *klass);
static void fo_property_wrap_option_class_init   (FoPropertyWrapOptionClass *klass);
static void fo_property_wrap_option_finalize     (GObject       *object);

static FoDatatype* fo_property_wrap_option_resolve_enum (const gchar *token,
                                                         FoContext   *context,
                                                         GError     **error);
static FoDatatype* fo_property_wrap_option_validate (FoDatatype *datatype,
                                                     FoContext  *context,
                                                     GError    **error);

static const gchar class_name[] = "wrap-option";
static gpointer parent_class;
static FoDatatype *enum_no_wrap;
static FoDatatype *enum_wrap;

/**
 * fo_property_wrap_option_get_type:
 * 
 * Register the #FoPropertyWrapOption type if not already registered and
 * return its #GType value.
 * 
 * Return value: #GType of #FoPropertyWrapOption.
 **/
GType
fo_property_wrap_option_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoPropertyWrapOptionClass),
        (GBaseInitFunc) fo_property_wrap_option_base_init,
        (GBaseFinalizeFunc) fo_property_wrap_option_base_finalize,
        (GClassInitFunc) fo_property_wrap_option_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoPropertyWrapOption),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_property_wrap_option_init,
	NULL		/* value_table */
      };

      object_type = g_type_register_static (FO_TYPE_PROPERTY,
                                            class_name,
                                            &object_info, 0);
    }

  return object_type;
}

/**
 * fo_property_wrap_option_init:
 * @wrap_option: #FoPropertyWrapOption object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoPropertyWrapOption.
 **/
void
fo_property_wrap_option_init (FoPropertyWrapOption *wrap_option)
{
  FO_PROPERTY (wrap_option)->value =
    g_object_ref (enum_wrap);
}

/**
 * fo_property_wrap_option_base_init:
 * @klass: #FoPropertyWrapOptionClass object to initialise.
 * 
 * Implements #GBaseInitFunc for #FoPropertyWrapOptionClass.
 **/
void
fo_property_wrap_option_base_init (FoPropertyWrapOptionClass *klass)
{
  enum_no_wrap =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_WRAP_OPTION),
                  "value",
                  FO_ENUM_WRAP_OPTION_NO_WRAP,
                  NULL);
  enum_wrap =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_WRAP_OPTION),
                  "value",
                  FO_ENUM_WRAP_OPTION_WRAP,
                  NULL);
}

/**
 * fo_property_wrap_option_base_finalize:
 * @klass: #FoPropertyWrapOptionClass object to finalise.
 * 
 * Implements #GBaseFinalizeFunc for #FoPropertyWrapOptionClass.
 **/
void
fo_property_wrap_option_base_finalize (FoPropertyWrapOptionClass *klass)
{
  g_object_unref (enum_no_wrap);
  g_object_unref (enum_wrap);
}

/**
 * fo_property_wrap_option_class_init:
 * @klass: #FoPropertyWrapOptionClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoPropertyWrapOptionClass.
 **/
void
fo_property_wrap_option_class_init (FoPropertyWrapOptionClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  FoPropertyClass *property_class = FO_PROPERTY_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = fo_property_wrap_option_finalize;

  property_class->is_inherited = TRUE;
  property_class->is_shorthand = FALSE;
  property_class->resolve_enum =
    fo_property_wrap_option_resolve_enum;
  property_class->validate =
    fo_property_wrap_option_validate;
  property_class->get_initial =
    fo_property_wrap_option_get_initial;
}

/**
 * fo_property_wrap_option_finalize:
 * @object: #FoPropertyWrapOption object to finalize.
 * 
 * Implements #GObjectFinalizeFunc for #FoPropertyWrapOption.
 **/
void
fo_property_wrap_option_finalize (GObject *object)
{
  FoPropertyWrapOption *wrap_option;

  wrap_option = FO_PROPERTY_WRAP_OPTION (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


/**
 * fo_property_wrap_option_new:
 * 
 * Creates a new #FoPropertyWrapOption initialized to default value.
 * 
 * Return value: the new #FoPropertyWrapOption.
 **/
FoProperty*
fo_property_wrap_option_new (void)
{
  FoProperty* wrap_option;

  wrap_option =
    FO_PROPERTY (g_object_new (fo_property_wrap_option_get_type (),
                               NULL));

  return wrap_option;
}

/**
 * fo_property_wrap_option_resolve_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_wrap_option_resolve_enum (const gchar *token,
                                      FoContext   *context,
                                      GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "no-wrap") == 0)
    {
      return g_object_ref (enum_no_wrap);
    }
  else if (strcmp (token, "wrap") == 0)
    {
      return g_object_ref (enum_wrap);
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   class_name,
		   token);
      return NULL;
    }
}

/**
 * fo_property_wrap_option_validate:
 * @datatype: #FoDatatype to be validated against allowed datatypes and
 *            values for current property.
 * @context:  #FoContext object from which to possibly inherit values.
 * @error:    Information about any error that has occurred.
 * 
 * Validates @datatype against allowed values.  Returns @datatype, a
 * replacement datatype value, or NULL if validation failed.
 * 
 * Return value: Valid datatype value or NULL.
 **/
FoDatatype*
fo_property_wrap_option_validate (FoDatatype *datatype,
                                  FoContext  *context,
                                  GError    **error)
{
  FoDatatype *new_datatype;
  GError     *tmp_error = NULL;
  gchar      *token;

  g_return_val_if_fail (datatype != NULL, NULL);
  g_return_val_if_fail (FO_IS_DATATYPE (datatype), NULL);
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (FO_IS_ENUM (datatype))
    {
      if ((G_ENUM_CLASS_TYPE (fo_enum_get_enum_class (datatype))
           == FO_TYPE_ENUM_WRAP_OPTION) &&
	  (g_enum_get_value (fo_enum_get_enum_class (datatype),
			     fo_enum_get_value (datatype)) != NULL))
	{
	  return datatype;
	}
      else
	{
	  gchar *datatype_sprintf = fo_object_sprintf (datatype);

	  g_set_error (error,
		       FO_FO_ERROR,
		       FO_FO_ERROR_DATATYPE,
		       _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		       class_name,
		       datatype_sprintf,
		       g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

	  g_object_unref (datatype);

	  g_free (datatype_sprintf);

	  return NULL;
	}
    }
  else if (FO_IS_STRING (datatype))
    {
      token = fo_string_get_value (datatype);

      new_datatype =
        fo_property_wrap_option_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else if (FO_IS_NAME (datatype))
    {
      token = fo_name_get_value (datatype);

      new_datatype =
        fo_property_wrap_option_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else
    {
      gchar *datatype_sprintf = fo_object_sprintf (datatype);

      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_DATATYPE,
		   _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		   class_name,
		   datatype_sprintf,
		   g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

      g_object_unref (datatype);

      g_free (datatype_sprintf);

      return NULL;
    }
}

/**
 * fo_property_wrap_option_get_initial:
 * 
 * Get an instance of the property with the correct initial value.
 * 
 * Return value: An instance of the property.
 **/
FoProperty*
fo_property_wrap_option_get_initial (void)
{
  static FoProperty *wrap_option = NULL;

  if (wrap_option == NULL)
    {
      wrap_option = fo_property_wrap_option_new ();
    }

  return wrap_option;
}
