/* Fo
 * fo-property-writing-mode.c: 'writing-mode' property
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-property-writing-mode.c,v 1.2 2003/03/21 14:41:46 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include <string.h>
#include "libfo/fo-utils.h"
#include "fo-context.h"
#include "datatype/fo-datatype.h"
#include "property/fo-property-private.h"
#include "property/fo-property-font-size.h"
#include "property/fo-property-writing-mode.h"

/* Inherited: TRUE */
/* Shorthand: FALSE */
/* writing-mode */
/* lr-tb | rl-tb | tb-rl | lr | rl | tb | inherit */
GType
fo_enum_writing_mode_get_type (void)
{
  static GType etype = 0;
  if (etype == 0)
    {
      static const GEnumValue values[] = {
        { FO_ENUM_WRITING_MODE_LR_TB, "FO_ENUM_WRITING_MODE_LR_TB", "lr-tb" },
        { FO_ENUM_WRITING_MODE_RL_TB, "FO_ENUM_WRITING_MODE_RL_TB", "rl-tb" },
        { FO_ENUM_WRITING_MODE_TB_RL, "FO_ENUM_WRITING_MODE_TB_RL", "tb-rl" },
        { 0, NULL, NULL }
      };
      etype = g_enum_register_static ("FoEnumWritingMode", values);
    }
  return etype;
}

struct _FoPropertyWritingMode
{
  FoProperty parent_instance;
};

struct _FoPropertyWritingModeClass
{
  FoPropertyClass parent_class;
};

static void fo_property_writing_mode_init         (FoPropertyWritingMode      *property_writing_mode);
static void fo_property_writing_mode_base_class_init   (FoPropertyWritingModeClass *klass);
static void fo_property_writing_mode_base_class_finalize   (FoPropertyWritingModeClass *klass);
static void fo_property_writing_mode_class_init   (FoPropertyWritingModeClass *klass);
static void fo_property_writing_mode_finalize     (GObject       *object);

static FoDatatype* fo_property_writing_mode_resolve_enum (const gchar *token,
                                                          FoContext   *context,
                                                          GError     **error);
static FoDatatype* fo_property_writing_mode_validate (FoDatatype *datatype,
                                                      FoContext  *context,
                                                      GError    **error);

static const gchar class_name[] = "writing-mode";
static gpointer parent_class;
static FoDatatype *enum_lr_tb;
static FoDatatype *enum_rl_tb;
static FoDatatype *enum_tb_rl;

/**
 * fo_property_writing_mode_get_type:
 * 
 * Register the #FoPropertyWritingMode type if not already registered and
 * return its #GType value.
 * 
 * Return value: #GType of #FoPropertyWritingMode.
 **/
GType
fo_property_writing_mode_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoPropertyWritingModeClass),
        (GBaseInitFunc) fo_property_writing_mode_base_class_init,
        (GBaseFinalizeFunc) fo_property_writing_mode_base_class_finalize,
        (GClassInitFunc) fo_property_writing_mode_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoPropertyWritingMode),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_property_writing_mode_init,
	NULL		/* value_table */
      };

      object_type = g_type_register_static (FO_TYPE_PROPERTY,
                                            class_name,
                                            &object_info, 0);
    }

  return object_type;
}

/**
 * fo_property_writing_mode_init:
 * @writing_mode: #FoPropertyWritingMode object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoPropertyWritingMode.
 **/
void
fo_property_writing_mode_init (FoPropertyWritingMode *writing_mode)
{
  FO_PROPERTY (writing_mode)->value =
    g_object_ref (enum_lr_tb);
}

/**
 * fo_property_writing_mode_base_class_init:
 * @klass: #FoPropertyWritingModeClass object to initialise
 * 
 * Implements #GBaseInitFunc for #FoPropertyWritingModeClass
 **/
void
fo_property_writing_mode_base_class_init (FoPropertyWritingModeClass *klass)
{
  enum_lr_tb =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_WRITING_MODE),
                  "value",
                  FO_ENUM_WRITING_MODE_LR_TB,
                  NULL);
  enum_rl_tb =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_WRITING_MODE),
                  "value",
                  FO_ENUM_WRITING_MODE_RL_TB,
                  NULL);
  enum_tb_rl =
    g_object_new (FO_TYPE_ENUM,
                  "enum-class",
                  g_type_class_ref (FO_TYPE_ENUM_WRITING_MODE),
                  "value",
                  FO_ENUM_WRITING_MODE_TB_RL,
                  NULL);
}

/**
 * fo_property_writing_mode_base_class_finalize:
 * @klass: #FoPropertyWritingModeClass object to finalise.
 * 
 * Implements #GBaseFinalizeFunc for #FoPropertyWritingModeClass.
 **/
void
fo_property_writing_mode_base_class_finalize (FoPropertyWritingModeClass *klass)
{
  g_object_unref (enum_lr_tb);
  g_object_unref (enum_rl_tb);
  g_object_unref (enum_tb_rl);
}

/**
 * fo_property_writing_mode_class_init:
 * @klass: #FoPropertyWritingModeClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoPropertyWritingModeClass.
 **/
void
fo_property_writing_mode_class_init (FoPropertyWritingModeClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  FoPropertyClass *property_class = FO_PROPERTY_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = fo_property_writing_mode_finalize;

  property_class->is_inherited = TRUE;
  property_class->is_shorthand = FALSE;
  property_class->resolve_enum =
    fo_property_writing_mode_resolve_enum;
  property_class->validate =
    fo_property_writing_mode_validate;
  property_class->get_initial =
    fo_property_writing_mode_get_initial;
}

/**
 * fo_property_writing_mode_finalize:
 * @object: #FoPropertyWritingMode object to finalize.
 * 
 * Implements #GObjectFinalizeFunc for #FoPropertyWritingMode.
 **/
void
fo_property_writing_mode_finalize (GObject *object)
{
  FoPropertyWritingMode *writing_mode;

  writing_mode = FO_PROPERTY_WRITING_MODE (object);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


/**
 * fo_property_writing_mode_new:
 * 
 * Creates a new #FoPropertyWritingMode initialized to default value.
 * 
 * Return value: the new #FoPropertyWritingMode.
 **/
FoProperty*
fo_property_writing_mode_new (void)
{
  FoProperty* writing_mode;

  writing_mode =
    FO_PROPERTY (g_object_new (fo_property_writing_mode_get_type (),
                               NULL));

  return writing_mode;
}

/**
 * fo_property_writing_mode_resolve_enum:
 * @token:   Token from the XML attribute value to be evaluated as an
 *           enumeration token.
 * @context: #FoContext object from which to possibly inherit values.
 * @error:   Information about any error that has occurred.
 * 
 * Compare @token against the enumeration tokens that are valid for the
 * current FO property.  If @token is valid, returns either an #FoEnum datatype
 * representing the enumeration token or a different datatype representing
 * the enumeration token's resolved value.  If @token is not valid,
 * sets @error and returns NULL.
 * 
 * Return value: Resolved enumeration value or NULL.
 **/
FoDatatype*
fo_property_writing_mode_resolve_enum (const gchar *token,
                                       FoContext   *context,
                                       GError     **error)
{
  g_return_val_if_fail (token != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (strcmp (token, "lr-tb") == 0)
    {
      return g_object_ref (enum_lr_tb);
    }
  else if (strcmp (token, "rl-tb") == 0)
    {
      return g_object_ref (enum_rl_tb);
    }
  else if (strcmp (token, "tb-rl") == 0)
    {
      return g_object_ref (enum_tb_rl);
    }
  else if (strcmp (token, "lr") == 0)
    {
      return g_object_ref (enum_lr_tb);
    }
  else if (strcmp (token, "rl") == 0)
    {
      return g_object_ref (enum_rl_tb);
    }
  else if (strcmp (token, "tb") == 0)
    {
      return g_object_ref (enum_tb_rl);
    }
  else if ((strcmp (token, "tb-lr") == 0) ||
	   (strcmp (token, "bt-lr") == 0) ||
	   (strcmp (token, "bt-rl") == 0) ||
	   (strcmp (token, "lr-bt") == 0) ||
	   (strcmp (token, "rl-bt") == 0) ||
	   (strcmp (token, "lr-alternating-rl-bt") == 0) ||
	   (strcmp (token, "lr-alternating-rl-tb") == 0) ||
	   (strcmp (token, "lr-inverting-rl-bt") == 0) ||
	   (strcmp (token, "lr-inverting-rl-tb") == 0) ||
	   (strcmp (token, "tb-lr-in-lr-pairs") == 0))
    {
      g_warning("Unsupported writing-mode: %s.  Using 'lr-tb'.", token);
      return g_object_ref (enum_lr_tb);
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_ENUMERATION_TOKEN,
		   _(fo_fo_error_messages[FO_FO_ERROR_ENUMERATION_TOKEN]),
		   class_name,
		   token);
      return NULL;
    }
}

/**
 * fo_property_writing_mode_validate:
 * @datatype: #FoDatatype to be validated against allowed datatypes and
 *            values for current property.
 * @context:  #FoContext object from which to possibly inherit values.
 * @error:    Information about any error that has occurred.
 * 
 * Validates @datatype against allowed values.  Returns @datatype, a
 * replacement datatype value, or NULL if validation failed.
 * 
 * Return value: Valid datatype value or NULL.
 **/
FoDatatype*
fo_property_writing_mode_validate (FoDatatype *datatype,
                                   FoContext  *context,
                                   GError    **error)
{
  FoDatatype *new_datatype;
  GError     *tmp_error = NULL;
  gchar      *token;

  g_return_val_if_fail (datatype != NULL, NULL);
  g_return_val_if_fail (FO_IS_DATATYPE (datatype), NULL);
  g_return_val_if_fail (context != NULL, NULL);
  g_return_val_if_fail (FO_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (FO_IS_ENUM (datatype))
    {
      if ((G_ENUM_CLASS_TYPE (fo_enum_get_enum_class (datatype))
           == FO_TYPE_ENUM_WRITING_MODE) &&
	  (g_enum_get_value (fo_enum_get_enum_class (datatype),
			     fo_enum_get_value (datatype)) != NULL))
	{
	  return datatype;
	}
      else
	{
	  g_set_error (error,
		       FO_FO_ERROR,
		       FO_FO_ERROR_DATATYPE,
		       _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		       class_name,
		       fo_object_sprintf (datatype),
		       g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

	  g_object_unref (datatype);

	  return NULL;
	}
    }
  else if (FO_IS_STRING (datatype))
    {
      token = fo_string_get_value (datatype);

      new_datatype =
        fo_property_writing_mode_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else if (FO_IS_NAME (datatype))
    {
      token = fo_name_get_value (datatype);

      new_datatype =
        fo_property_writing_mode_resolve_enum (token, context, &tmp_error);

      g_object_unref (datatype);

      if (tmp_error != NULL)
	{
	  g_propagate_error (error, tmp_error);
	  return NULL;
	}

      return new_datatype;
    }
  else
    {
      g_set_error (error,
		   FO_FO_ERROR,
		   FO_FO_ERROR_DATATYPE,
		   _(fo_fo_error_messages[FO_FO_ERROR_DATATYPE]),
		   class_name,
		   fo_object_sprintf (datatype),
		   g_type_name (G_TYPE_FROM_INSTANCE (datatype)));

      g_object_unref (datatype);

      return NULL;
    }
}

/**
 * fo_property_writing_mode_get_initial:
 * 
 * Get an instance of the property with the correct initial value.
 * 
 * Return value: An instance of the property.
 **/
FoProperty*
fo_property_writing_mode_get_initial (void)
{
  static FoProperty *writing_mode = NULL;

  if (writing_mode == NULL)
    {
      writing_mode = fo_property_writing_mode_new ();
    }

  return writing_mode;
}

/**
 * fo_property_writing_mode_to_bpd:
 * @writing_mode: #FoPropertyWritingMode
 * @error:        #GError
 * 
 * Determine the block-progression-direction for the current value of
 * @writing_mode
 * 
 * Return value: #FoEnumAreaDirection indicating block-progression-direction
 **/
FoEnumAreaDirection
fo_property_writing_mode_to_bpd (FoProperty  *writing_mode,
				 GError     **error)
{
  gint writing_mode_value;
  FoEnumAreaDirection bpd = FO_ENUM_AREA_DIRECTION_UNKNOWN;

  g_return_val_if_fail (writing_mode != NULL,
			FO_ENUM_AREA_DIRECTION_UNKNOWN);
  g_return_val_if_fail (FO_IS_PROPERTY_WRITING_MODE (writing_mode),
			FO_ENUM_AREA_DIRECTION_UNKNOWN);
  g_return_val_if_fail (error == NULL || *error == NULL,
			FO_ENUM_AREA_DIRECTION_UNKNOWN);

  writing_mode_value =
    fo_enum_get_value (writing_mode->value);

  if (writing_mode_value == FO_ENUM_WRITING_MODE_LR_TB)
    {
      bpd = FO_ENUM_AREA_DIRECTION_TB;
    }
  else if (writing_mode_value == FO_ENUM_WRITING_MODE_RL_TB)
    {
      bpd = FO_ENUM_AREA_DIRECTION_TB;
    }
  else if (writing_mode_value == FO_ENUM_WRITING_MODE_TB_RL)
    {
      bpd = FO_ENUM_AREA_DIRECTION_RL;
    }
  else
    {
      g_assert_not_reached ();
    }

  return bpd;
}

/**
 * fo_property_writing_mode_to_ipd:
 * @writing_mode: #FoPropertyWritingMode
 * @error:        #GError
 * 
 * Determine the inline-progression-direction for the current value of
 * @writing_mode
 * 
 * Return value: #FoEnumAreaDirection indicating inline-progression-direction
 **/
FoEnumAreaDirection
fo_property_writing_mode_to_ipd (FoProperty  *writing_mode,
				 GError     **error)
{
  gint writing_mode_value;
  FoEnumAreaDirection ipd = FO_ENUM_AREA_DIRECTION_UNKNOWN;

  g_return_val_if_fail (writing_mode != NULL,
			FO_ENUM_AREA_DIRECTION_UNKNOWN);
  g_return_val_if_fail (FO_IS_PROPERTY_WRITING_MODE (writing_mode),
			FO_ENUM_AREA_DIRECTION_UNKNOWN);
  g_return_val_if_fail (error == NULL || *error == NULL,
			FO_ENUM_AREA_DIRECTION_UNKNOWN);

  writing_mode_value =
    fo_enum_get_value (writing_mode->value);

  if (writing_mode_value == FO_ENUM_WRITING_MODE_LR_TB)
    {
      ipd = FO_ENUM_AREA_DIRECTION_LR;
    }
  else if (writing_mode_value == FO_ENUM_WRITING_MODE_RL_TB)
    {
      ipd = FO_ENUM_AREA_DIRECTION_RL;
    }
  else if (writing_mode_value == FO_ENUM_WRITING_MODE_TB_RL)
    {
      ipd = FO_ENUM_AREA_DIRECTION_TB;
    }
  else
    {
      g_assert_not_reached ();
    }

  return ipd;
}

/**
 * fo_property_writing_mode_to_sd:
 * @writing_mode: #FoPropertyWritingMode
 * @error:        #GError
 * 
 * Determine the shift-direction for the current value of
 * @writing_mode
 * 
 * Return value: #FoEnumAreaDirection indicating shift-direction
 **/
FoEnumAreaDirection
fo_property_writing_mode_to_sd (FoProperty  *writing_mode,
				 GError     **error)
{
  gint writing_mode_value;
  FoEnumAreaDirection sd = FO_ENUM_AREA_DIRECTION_UNKNOWN;

  g_return_val_if_fail (writing_mode != NULL,
			FO_ENUM_AREA_DIRECTION_UNKNOWN);
  g_return_val_if_fail (FO_IS_PROPERTY_WRITING_MODE (writing_mode),
			FO_ENUM_AREA_DIRECTION_UNKNOWN);
  g_return_val_if_fail (error == NULL || *error == NULL,
			FO_ENUM_AREA_DIRECTION_UNKNOWN);

  writing_mode_value =
    fo_enum_get_value (writing_mode->value);

  if (writing_mode_value == FO_ENUM_WRITING_MODE_LR_TB)
    {
      sd = FO_ENUM_AREA_DIRECTION_BT;
    }
  else if (writing_mode_value == FO_ENUM_WRITING_MODE_RL_TB)
    {
      sd = FO_ENUM_AREA_DIRECTION_BT;
    }
  else if (writing_mode_value == FO_ENUM_WRITING_MODE_TB_RL)
    {
      sd = FO_ENUM_AREA_DIRECTION_LR;
    }
  else
    {
      g_assert_not_reached ();
    }

  return sd;
}

