/* Fo
 * result-to-fo.c: Convert result tree into formatting object tree
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: result-to-fo.c,v 1.23 2005/01/03 21:01:38 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */


#include <string.h>
#include "libfo/fo-utils.h"
#include "fo-context-util.h"
#include "fo/fo-fo.h"
#include "fo/fo-fo-private.h"
#include "fo-tree.h"
#include "fo-area-tree.h"
/* FIXME: Shouldn't be necessary to include -private.h file*/
/*#include "fo-area-area-private.h"*/
#include "fo-all-fo.h"
#include "fo-property-eval.h"
#include "fo-all-property.h"
#include <libxml/xmlversion.h>
#include <libxml/xmlmemory.h>
#include <libxml/debugXML.h>
#include <libxml/HTMLtree.h>
#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>
#include "result-to-fo.h"

#define XSL_FO_NAMESPACE \
    "http://www.w3.org/1999/XSL/Format"

typedef GType (*FoTypeFunc) (void);

typedef FoFo* (*FoXMLNodeToFoNode) (xmlNodePtr node,
				    GError **error);


const char *fo_result_to_fo_error_messages[] = {
  N_("Failed"),
  N_("Warning"),
  N_("Unexpected node type"),
  N_("Unknown 'fo:' element type: %s"),
  N_("Unsupported property: %s")
};

/**
 * fo_result_to_fo_error_quark:
 * 
 * 
 * 
 * Return value: #GQuark identifying errors converting result tree to FO tree.
 **/
GQuark
fo_result_to_fo_error_quark (void)
{
  static GQuark quark = 0;
  if (quark == 0)
    quark = g_quark_from_static_string ("fo-result-to-fo-error-quark");
  return quark;
}

typedef struct _xslAttrListIterator xslAttrListIterator;
typedef xslAttrListIterator *xslAttrListIteratorPtr;
struct _xslAttrListIterator {
  xmlAttrPtr cur;
};

/**
 * xslAttrListGetIterator:
 * @node: 
 * 
 * 
 * 
 * Return value: 
 **/
static xslAttrListIteratorPtr
xslAttrListGetIterator (xmlNodePtr node)
{
  xslAttrListIteratorPtr iterator;

  if (node->properties)
    {
      iterator = (xslAttrListIteratorPtr) xmlMalloc (sizeof (xslAttrListIterator));

      iterator->cur = node->properties;
    }
  else
    {
      iterator = NULL;
    }

  return (iterator);
}

/**
 * xslAttrListIteratorNext:
 * @iterator: 
 * 
 * 
 * 
 * Return value: 
 **/
static gboolean
xslAttrListIteratorNext (xslAttrListIteratorPtr iterator)
{
  if (iterator->cur != NULL)
    {
      iterator->cur = iterator->cur->next;
    }
  return iterator->cur ? TRUE : FALSE;
}

/**
 * xslAttrListIteratorName:
 * @iterator: 
 * 
 * 
 * 
 * Return value: 
 **/
static xmlChar *
xslAttrListIteratorName (xslAttrListIteratorPtr iterator)
{
  return (xmlChar*) iterator->cur->name;
}

/**
 * xslAttrListIteratorValue:
 * @iterator: 
 * 
 * 
 * 
 * Return value: 
 **/
static xmlChar *
xslAttrListIteratorValue (xslAttrListIteratorPtr iterator)
{
  return (xmlChar*) iterator->cur->children->content;
}

/**
 * xslAttrListIteratorDestroy:
 * @iterator: 
 * 
 * 
 **/
static void
xslAttrListIteratorDestroy (xslAttrListIteratorPtr iterator)
{
  xmlFree (iterator);
}

static FoFo*
fo_xml_element_node_to_fo_node (xmlNodePtr node,
				GError **error)
{
  FoFo *fo_node = NULL;
  gboolean ok = FALSE;
  xmlChar *name;

  g_return_val_if_fail (node != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  name = (xmlChar*) node->name;

  switch (name[0]) {
  case 'b':
    switch (name[1]) {
    case 'a':
      if (strcmp (name, "basic-link") == 0) {
	fo_node = fo_basic_link_new ();
	ok = TRUE;
      }
      break; /* end 'ba' */
    case 'i':
      if (strcmp (name, "bidi-override") == 0) {
	fo_node = fo_bidi_override_new ();
	ok = TRUE;
      }
      break; /* end 'bi' */
    case 'l':
      if (strcmp (name, "block") == 0) {
	fo_node = fo_block_new ();
	ok = TRUE;
      } else if (strcmp (name, "block-container") == 0) {
	fo_node = fo_block_container_new ();
	ok = TRUE;
      }
      break; /* end 'bl' */
    }
    break; /* end 'b' */
  case 'c':
    if (name[1] == 'h') {
      if (strcmp (name, "character") == 0) {
	fo_node = fo_character_new ();
	ok = TRUE;
      }
    } else if (name[1] == 'o') {
      if (strcmp (name, "color-profile") == 0) {
	fo_node = fo_color_profile_new ();
	ok = TRUE;
      } else if (strcmp (name, "conditional-page-master-reference") == 0) {
	fo_node = fo_conditional_page_master_reference_new ();
	ok = TRUE;
      }
    }
    break; /* end 'c' */
  case 'd':
    if (strcmp (name, "declarations") == 0) {
      fo_node = fo_declarations_new ();
      ok = TRUE;
    }
    break; /* end 'd' */
  case 'e':
    if (strcmp (name, "external-graphic") == 0) {
      fo_node = fo_external_graphic_new ();
      ok = TRUE;
    }
    break; /* end 'e' */
  case 'f':
    if ((name[1] == 'l') && (name[2] == 'o')) {
      if (strcmp (name, "flow") == 0) {
	fo_node = fo_flow_new ();
	fo_node->flow = g_object_ref (fo_node);
	ok = TRUE;
      } else if (strcmp (name, "float") == 0) {
	fo_node = fo_float_new ();
	ok = TRUE;
      }
    } else if (name[1] == 'o') {
      if (strncmp (name, "footnote", 8) == 0) {
	if (name[8] == '\0') {
	  fo_node = fo_footnote_new ();
	  ok = TRUE;
	} else if (strcmp (&name[8], "-body") == 0) {
	  fo_node = fo_footnote_body_new ();
	  ok = TRUE;
	}
      }
    }
    break; /* end 'f' */
  case 'i':
    if (name[1] == 'n') {
      if (strcmp (name, "initial-property-set") == 0) {
	fo_node = fo_initial_property_set_new ();
	ok = TRUE;
      } else if (strncmp (name, "inline", 6) == 0) {
	if (name[6] == '\0') {
	  fo_node = fo_inline_new ();
	  ok = TRUE;
	} else if (strcmp (&name[6], "-container") == 0) {
	  fo_node = fo_inline_container_new ();
	  ok = TRUE;
	}
      } else if (strcmp (name, "instream-foreign-object") == 0) {
	fo_node = fo_instream_foreign_object_new ();
	ok = TRUE;
      }
    }
    break; /* end 'i' */
  case 'l':
    switch (name[1]) {
    case 'a':
      if (strcmp (name, "layout-master-set") == 0) {
	fo_node = fo_layout_master_set_new ();
	ok = TRUE;
      }
      break; /* end 'la' */
    case 'e':
      if (strcmp (name, "leader") == 0) {
	fo_node = fo_leader_new ();
	ok = TRUE;
      }
      break; /* end 'le' */
    case 'i':
      if (strncmp (name, "list-", 5) == 0) {
	if (strcmp (&name[5], "block") == 0) {
	  fo_node = fo_list_block_new ();
	  ok = TRUE;
	} else if (strcmp (&name[5], "item") == 0) {
	  fo_node = fo_list_item_new ();
	  ok = TRUE;
	} else if (strcmp (&name[5], "item-body") == 0) {
	  fo_node = fo_list_item_body_new ();
	  ok = TRUE;
	} else if (strcmp (&name[5], "item-label") == 0) {
	  fo_node = fo_list_item_label_new ();
	  ok = TRUE;
	}
      }
      break; /* end 'li' */
    }
    break; /* end 'l' */
  case 'm':
    if (name[1] == 'a') {
      if (strcmp (name, "marker") == 0) {
	fo_node = fo_marker_new ();
	ok = TRUE;
      }
    } else if (strncmp (name, "multi-", 6) == 0) {
      if (strcmp (&name[6], "case") == 0) {
	fo_node = fo_multi_case_new ();
	ok = TRUE;
      } else if (strcmp (&name[6], "properties") == 0) {
	fo_node = fo_multi_properties_new ();
	ok = TRUE;
      } else if (strcmp (&name[6], "property-set") == 0) {
	fo_node = fo_multi_property_set_new ();
	ok = TRUE;
      } else if (strcmp (&name[6], "switch") == 0) {
	fo_node = fo_multi_switch_new ();
	ok = TRUE;
      } else if (strcmp (&name[6], "toggle") == 0) {
	fo_node = fo_multi_toggle_new ();
	ok = TRUE;
      }
    }
    break; /* end 'm' */
  case 'p':
    if (strncmp (name, "page-", 5) == 0) {
      if (strncmp (&name[5], "number", 6) == 0) {
	if (name[11] == '\0') {
	  fo_node = fo_page_number_new ();
	  ok = TRUE;
	} else if (strcmp (&name[11], "-citation") == 0) {
	  fo_node = fo_page_number_citation_new ();
	  ok = TRUE;
	}
      } else if (strncmp (&name[5], "sequence", 8) == 0) {
	if (name[13] == '\0') {
	  fo_node = fo_page_sequence_new ();
	  ok = TRUE;
	} else if (strcmp (&name[13], "-master") == 0) {
	  fo_node = fo_page_sequence_master_new ();
	  ok = TRUE;
	}
      }
    }
    break; /* end 'p' */
  case 'r':
    switch (name[1]) {
    case 'e':
      if (strncmp (name, "region-", 7) == 0) {
	if (strcmp (&name[7], "after") == 0) {
	  fo_node = fo_region_after_new ();
	  ok = TRUE;
	} else if (strcmp (&name[7], "before") == 0) {
	  fo_node = fo_region_before_new ();
	  ok = TRUE;
	} else if (strcmp (&name[7], "body") == 0) {
	  fo_node = fo_region_body_new ();
	  ok = TRUE;
	} else if (strcmp (&name[7], "end") == 0) {
	  fo_node = fo_region_end_new ();
	  ok = TRUE;
	} else if (strcmp (&name[7], "start") == 0) {
	  fo_node = fo_region_start_new ();
	  ok = TRUE;
	}
      } else if (strncmp (name, "repeatable-page-master-", 23) == 0) {
	if (strcmp (&name[23], "alternatives") == 0) {
	  fo_node = fo_repeatable_page_master_alternatives_new ();
	  ok = TRUE;
	} else if (strcmp (&name[23], "reference") == 0) {
	  fo_node = fo_repeatable_page_master_reference_new ();
	  ok = TRUE;
	}
      } else if (strcmp (name, "retrieve-marker") == 0) {
	fo_node = fo_retrieve_marker_new ();
	ok = TRUE;
      }
      break; /* end 're' */
    case 'o':
      if (strcmp (name, "root") == 0) {
	fo_node = fo_root_new ();
	ok = TRUE;
      }
      break; /* end 'ro' */
    }
    break; /* end 'r' */
  case 's':
    if (name[1] == 'i') {
      if (strcmp (name, "simple-page-master") == 0) {
	fo_node = fo_simple_page_master_new ();
	ok = TRUE;
      } else if (strcmp (name, "single-page-master-reference") == 0) {
	fo_node = fo_single_page_master_reference_new ();
	ok = TRUE;
      }
    } else if (name[1] == 't') {
      if (strcmp (name, "static-content") == 0) {
	fo_node = fo_static_content_new ();
	ok = TRUE;
      }
    }
    break; /* end 's' */
  case 't':
    if (strncmp (name, "table", 5) == 0) {
      if (name[5] == '\0') {
	fo_node = fo_table_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-and-caption") == 0) {
	fo_node = fo_table_and_caption_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-body") == 0) {
	fo_node = fo_table_body_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-caption") == 0) {
	fo_node = fo_table_caption_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-cell") == 0) {
	fo_node = fo_table_cell_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-column") == 0) {
	fo_node = fo_table_column_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-footer") == 0) {
	fo_node = fo_table_footer_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-header") == 0) {
	fo_node = fo_table_header_new ();
	ok = TRUE;
      } else if (strcmp (&name[5], "-row") == 0) {
	fo_node = fo_table_row_new ();
	ok = TRUE;
      } /* end 'table' */
    } else if (strcmp (name, "title") == 0) {
      fo_node = fo_title_new ();
      ok = TRUE;
    }
    break; /* end 't' */
  case 'w':
    if (strcmp (name, "wrapper") == 0) {
      fo_node = fo_wrapper_new ();
      ok = TRUE;
    }
    break; /* end 'w' */
  }    

  if (!ok)
    {
      g_set_error (error,
		   FO_RESULT_TO_FO_ERROR,
		   FO_RESULT_TO_FO_ERROR_UNKNOWN_FO,
		   _(fo_result_to_fo_error_messages[FO_RESULT_TO_FO_ERROR_UNKNOWN_FO]),
		   name);
      return NULL;
    }
  else
    {
      fo_node->element = node;
      return fo_node;
    }
}

/**
 * fo_xml_node_to_fo_subtree:
 * @node: 
 * @parent_context: 
 * @prop_eval_hash: 
 * @parent_fo: 
 * @error: 
 * 
 * 
 **/
static void
fo_new_xml_to_fo_tree (xmlNodePtr node,
		       FoFo *parent_fo,
		       GError **error)
{
  FoFo *fo_node = NULL;
  FoFo *fo_text = NULL;
  GError *tmp_error = NULL;

  g_return_if_fail (node != NULL);
  g_return_if_fail (error == NULL || *error == NULL);

  switch (node->type)
    {
    case XML_ELEMENT_NODE:
      if ((node->ns != NULL) &&
	  (strcmp (node->ns->href, XSL_FO_NAMESPACE) == 0 ))
	{
	  fo_node = fo_xml_element_node_to_fo_node (node,
						    &tmp_error);

	  if (tmp_error != NULL)
	    {
	      g_propagate_error (error, tmp_error);
	      return;
	    }

#if defined(LIBFO_DEBUG) && 0
	  g_message ("new_xml_to_fo_tree:: fo_node: %s",
		     fo_object_sprintf (FO_OBJECT (fo_node)));
#endif

	  if (fo_node)
	    {
	      fo_node_append (FO_NODE (parent_fo),
			      FO_NODE (fo_node));
	    }

	  node = node->children;

	  while (node)
	    {
	      /* Get the next node now, since current node may be PI or
		 comment that is unlinked and freed in following call to
		 fo_xml_node_to_fo_node() */
	      xmlNodePtr next_node = node->next;

	      fo_new_xml_to_fo_tree (node,
				     fo_node,
				     &tmp_error);

	      if (tmp_error != NULL)
		{
		  g_propagate_error (error, tmp_error);
		  return;
		}

	      node = next_node;
	    }
	}
      break;
    case XML_ATTRIBUTE_NODE:
      g_error (_("Unexpected node type: ATTRIBUTE"));
      break;
    case XML_TEXT_NODE:
      {
	xmlChar *text = xmlNodeGetContent (node);

	if (fo_fo_get_allow_mixed_content (parent_fo))
	  {
	    fo_text = fo_text_new ();
	    fo_text_set_value (fo_text, text);
	    fo_fo_update_from_context (fo_text,
				       fo_fo_get_context (parent_fo));
	    fo_node_append (FO_NODE (parent_fo),
			    FO_NODE (fo_text));
	  }
	else
	  {
	    xmlChar *text_ptr = text;
	    gboolean whitespace_only = TRUE;

	    while (*text_ptr)
	      {
		if (!g_ascii_isspace (*text_ptr++))
		  {
		    whitespace_only = FALSE;
		    break;
		  }
	      }

	    if (!whitespace_only)
	      {
		g_error (_("Text node should contain only whitespace characters: \"%s\""), text);
	      }
	    xmlUnlinkNode (node);
	    xmlFreeNode (node);
	  }

	xmlFree (text);
      }
      break;
    case XML_CDATA_SECTION_NODE:
      g_error (_("Unexpected node type: CDATA SECTION"));
      break;
    case XML_ENTITY_REF_NODE:
      g_error (_("Unexpected node type: ENTITY REF"));
      break;
    case XML_ENTITY_NODE:
      g_error (_("Unexpected node type: ENTITY"));
      break;
    case XML_PI_NODE:
      /* Fall through */
    case XML_COMMENT_NODE:
      /* Unlink from the node tree */
      xmlUnlinkNode (node);
      /* Free the node */
      xmlFreeNode (node);
      break;
    case XML_DOCUMENT_NODE:
	  node = node->children;

	  while (node)
	    {
	      /* Get the next node now, since current node may be PI or
		 comment that is unlinked and freed in following call to
		 fo_xml_node_to_fo_node() */
	      xmlNodePtr next_node = node->next;

	      fo_new_xml_to_fo_tree (node,
				     parent_fo,
				     &tmp_error);
	      if (tmp_error != NULL)
		{
		  g_propagate_error (error, tmp_error);
		  return;
		}

	      node = next_node;
	    }
      /*
      g_error (_("Unexpected node type: DOCUMENT"));
      */
      break;
    case XML_DOCUMENT_TYPE_NODE:
      g_error (_("Unexpected node type: DOCUMENT TYPE"));
      break;
    case XML_DOCUMENT_FRAG_NODE:
      g_error (_("Unexpected node type: DOCUMENT FRAG"));
      break;
    case XML_NOTATION_NODE:
      g_error (_("Unexpected node type: NOTATION"));
      break;
    case XML_HTML_DOCUMENT_NODE:
      g_error (_("Unexpected node type: HTML DOCUMENT"));
      break;
    case XML_DTD_NODE:
      g_error (_("Unexpected node type: DTD"));
      break;
    case XML_ELEMENT_DECL:
      g_error (_("Unexpected node type: ELEMENT DECL"));
      break;
    case XML_ATTRIBUTE_DECL:
      g_error (_("Unexpected node type: ATTRIBUTE DECL"));
      break;
    case XML_ENTITY_DECL:
      g_error (_("Unexpected node type: ENTITY DECL"));
      break;
    case XML_NAMESPACE_DECL:
      g_error (_("Unexpected node type: NAMESPACE DECL"));
      break;
    case XML_XINCLUDE_START:
      g_error (_("Unexpected node type: XINCLUDE START"));
      break;
    case XML_XINCLUDE_END:
      g_error (_("Unexpected node type: XINCLUDE END"));
      break;
    case XML_DOCB_DOCUMENT_NODE:
      g_error (_("Unexpected node type: DOCB DOCUMENT"));
      break;
    }
}

/**
 * fo_fo_resolve_property_attributes:
 * @fo_node:      #FoFo
 * @prop_context: Context within which to resolve property expressions
 * 
 * Every #FoFo object was created from a result tree element that is
 * in the XSL FO namespace.  The object's specified property values
 * are created from the result tree element's attributes.
 *
 * This function evaluates each of the property attributes of the
 * result tree element for @fo_node.
 * 
 * Return value: FALSE if completed successfully, TRUE otherwise
 **/
gboolean
fo_fo_resolve_property_attributes_default (FoNode     *fo_node,
					  gpointer    data)
{
  FoPropertyResolveContext *prop_context = (FoPropertyResolveContext *) data;
  xmlNodePtr element;
  xslAttrListIteratorPtr iterator;
  FoProperty *font_size = NULL;
  FoFo *fo_fo;
  GSList *node_properties = NULL;
  FoContext *current_context;
  FoContext *parent_context;
  GError *error = NULL;
  FoArea *reference_area = prop_context->reference_area;
  GHashTable *prop_eval_hash = prop_context->prop_eval_hash;

  g_return_val_if_fail (FO_IS_FO (fo_node), TRUE);
  g_return_val_if_fail (prop_eval_hash != NULL, TRUE);

#if defined(LIBFO_DEBUG) && 0
  g_message ("fo_resolve_property_attributes:: fo_node: %s; reference: %s",
	     fo_object_debug_sprintf (fo_node),
	     fo_object_debug_sprintf (reference_area));
#endif

  fo_fo = FO_FO (fo_node);

  element = fo_fo->element;
  parent_context =
    fo_fo_get_context (FO_FO (fo_node_parent (FO_NODE (fo_fo))));

  if (FO_IS_TABLE_CELL (fo_fo))
    {
      xmlAttrPtr column_number_attr = NULL;
      xmlAttrPtr number_columns_spanned_attr = NULL;
      xmlAttrPtr number_rows_spanned_attr = NULL;

      column_number_attr = xmlHasProp (element, "column-number");
      if (column_number_attr)
	{
	  FoTypeFunc type_func =
	    g_hash_table_lookup (prop_eval_hash,
				 "column-number");

	  FoPropertyClass *property_class =
	    g_type_class_ref (type_func());

	  const xmlChar *property_expression =
	    xmlNodeGetContent ((xmlNodePtr) column_number_attr);

	  FoProperty *column_number =
	    fo_property_new_from_expr (property_class,
				       property_expression,
				       parent_context,
				       NULL,
				       fo_fo,
				       &error);

	  if (error != NULL)
	    {
	      fo_object_log_error (FO_OBJECT (fo_fo),
				   &error);
	    }

	  node_properties = g_slist_prepend (node_properties,
					     (gpointer) column_number);
	  fo_table_cell_set_column_number (fo_fo,
					   column_number);
	}

      number_columns_spanned_attr =
	xmlHasProp (element, "number-columns-spanned");
      if (number_columns_spanned_attr)
	{
	  FoTypeFunc type_func =
	    g_hash_table_lookup (prop_eval_hash,
				 "number-columns-spanned");

	  FoPropertyClass *property_class =
	    g_type_class_ref (type_func());

	  const xmlChar *property_expression =
	    xmlNodeGetContent ((xmlNodePtr) number_columns_spanned_attr);

	  FoProperty *number_columns_spanned =
	    fo_property_new_from_expr (property_class,
				       property_expression,
				       parent_context,
				       NULL,
				       fo_fo,
				       &error);

	  if (error != NULL)
	    {
	      fo_object_log_error (FO_OBJECT (fo_fo),
				   &error);
	    }

	  node_properties = g_slist_prepend (node_properties,
					     (gpointer) number_columns_spanned);
	  fo_table_cell_set_number_columns_spanned (fo_fo,
						    number_columns_spanned);
	}

      number_rows_spanned_attr = xmlHasProp (element, "number-rows-spanned");
      if (number_rows_spanned_attr)
	{
	  FoTypeFunc type_func =
	    g_hash_table_lookup (prop_eval_hash,
				 "number-rows-spanned");

	  FoPropertyClass *property_class =
	    g_type_class_ref (type_func());

	  const xmlChar *property_expression =
	    xmlNodeGetContent ((xmlNodePtr) number_rows_spanned_attr);

	  FoProperty *number_rows_spanned =
	    fo_property_new_from_expr (property_class,
				       property_expression,
				       parent_context,
				       NULL,
				       fo_fo,
				       &error);

	  if (error != NULL)
	    {
	      fo_object_log_error (FO_OBJECT (fo_fo),
				   &error);
	    }

	  node_properties = g_slist_prepend (node_properties,
					     (gpointer) number_rows_spanned);
	  fo_table_cell_set_number_rows_spanned (fo_fo,
						 number_rows_spanned);
	}
      fo_table_cell_resolve_number_columns_spanned (fo_fo);
      fo_table_cell_resolve_number_rows_spanned (fo_fo);
      fo_table_cell_resolve_column_number (fo_fo,
					   &error);
      if (error != NULL)
	{
	  fo_object_log_error (FO_OBJECT (fo_fo),
			       &error);
	}

      fo_table_cell_resolve_column (fo_fo);
    }

  if (xmlHasProp (element, "font-size"))
    {
      xmlAttrPtr font_size_attr = xmlHasProp (element, "font-size");

      FoTypeFunc type_func =
	g_hash_table_lookup (prop_eval_hash,
			     "font-size");

      FoPropertyClass *property_class =
	g_type_class_ref (type_func());

      font_size =
	fo_property_new_from_expr (property_class,
				   xmlNodeGetContent ((xmlNodePtr) font_size_attr),
				   parent_context,
				   NULL,
				   fo_fo,
				   &error);

      if (error != NULL)
	{
	  fo_object_log_error (FO_OBJECT (fo_fo),
			       &error);
	}
      else
	{
	  node_properties = g_slist_prepend (node_properties,
					     (gpointer) font_size);
	}
    }


  iterator = xslAttrListGetIterator (element);


  while (iterator)
    {
      FoProperty *property = NULL;

      if (strcmp (xslAttrListIteratorName (iterator), "font-size") == 0 ||
	  (FO_IS_TABLE_CELL (fo_fo) &&
	   (strcmp (xslAttrListIteratorName (iterator),
		    "column-number") == 0 ||
	    strcmp (xslAttrListIteratorName (iterator),
		    "number-columns-spanned") == 0 ||
	    strcmp (xslAttrListIteratorName (iterator),
		    "number-rows-spanned") == 0)))
	{
	  if (xslAttrListIteratorNext (iterator))
	    {
	      continue;
	    }
	  else
	    {
	      break;
	    }
	}

      FoPropertyClass *property_class = NULL;
      FoTypeFunc type_func =
	g_hash_table_lookup (prop_eval_hash,
			     xslAttrListIteratorName (iterator));

      if (type_func != NULL)
	{
	  property_class =
	    g_type_class_ref (type_func());

	  property =
	    fo_property_new_from_expr (property_class,
				       xslAttrListIteratorValue (iterator),
				       parent_context,
				       font_size ? font_size : NULL,
				       fo_fo,
				       &error);
	  if (error != NULL)
	    {
	      fo_object_log_error (FO_OBJECT (fo_fo),
				   &error);
	    }

	  g_type_class_unref (property_class);
	  node_properties = g_slist_prepend (node_properties,
					     (gpointer) property);
	}
      else
	{
	  GError *unsupported_property_error =
	    g_error_new (FO_RESULT_TO_FO_ERROR,
			 FO_RESULT_TO_FO_ERROR_UNSUPPORTED_PROPERTY,
			 _(fo_result_to_fo_error_messages[FO_RESULT_TO_FO_ERROR_UNSUPPORTED_PROPERTY]),
			 xslAttrListIteratorName (iterator));

	  fo_object_log_warning (FO_OBJECT (fo_fo),
				 &unsupported_property_error);
	}

    if (!xslAttrListIteratorNext (iterator))
      break;
  }

  xslAttrListIteratorDestroy (iterator);

  current_context = fo_context_new ();
  if (node_properties != NULL)
    {
      fo_context_update_from_slist (current_context, node_properties);
      g_slist_free (node_properties);
    }

  fo_context_util_spaces_resolve (current_context, reference_area);
  fo_context_util_length_ranges_resolve (current_context, reference_area);
  fo_fo_validate (fo_fo,
		  current_context,
		  parent_context,
		  &error);
  if (error != NULL)
    {
      fo_object_log_error (FO_OBJECT (fo_fo),
			   &error);
    }
  fo_context_merge (current_context, parent_context);
  fo_fo_set_context (fo_fo, current_context);

  return FALSE;
}

/**
 * fo_xml_doc_to_fo_and_area_trees:
 * @res: 
 * @fo_doc: 
 * @fo_tree: 
 * @area_tree: 
 * @debug_level:
 * @error: 
 * 
 * 
 **/
void
fo_xml_doc_to_fo_and_area_trees (xmlDocPtr res,
				 FoDoc *fo_doc,
				 FoFo **fo_tree,
				 FoArea **area_tree,
				 gint debug_level,
				 GError **error)
{
  FoFo *new_fo_tree = fo_tree_new ();
  FoArea *new_area_tree;
  FoContext *new_fo_context;
  GHashTable *prop_eval_hash = NULL;
  FoArea *new_area = NULL;
  GError *tmp_error = NULL;

  /* Check the xmlDocPtr isn't null */
  g_return_if_fail (res != NULL);
  /* Check the xmlDocPtr has a document element */
  g_return_if_fail (res->children != NULL);
  g_return_if_fail (error == NULL || *error == NULL);

  prop_eval_hash = fo_property_eval_init_hash (prop_eval_hash);
  new_fo_context = fo_context_new ();
  fo_context_initialize (new_fo_context);
  new_area_tree = fo_area_tree_new ();
  new_fo_tree->areas = g_list_append (new_fo_tree->areas, new_area_tree);
  fo_fo_set_context (new_fo_tree, new_fo_context);
  new_fo_tree->tree = new_fo_tree;

  fo_new_xml_to_fo_tree ((xmlNodePtr) res,
			 new_fo_tree,
			 NULL);

  fo_node_traverse (FO_NODE (new_fo_tree),
		    G_POST_ORDER,
		    G_TRAVERSE_NON_LEAFS,
		    -1,
		    (FoNodeTraverseFunc) fo_fo_validate_content,
		    NULL);
#if defined(LIBFO_DEBUG) && 0
  fo_object_debug_dump (new_fo_tree, 0);
#endif
  fo_fo_children_properties_resolve (new_fo_tree,
				     new_area_tree,
				     &new_area,
				     prop_eval_hash,
				     fo_doc,
				     debug_level,
				     &tmp_error);

  if (tmp_error != NULL)
    {
      fo_object_log_or_propagate_error (FO_OBJECT (new_fo_tree),
					error,
					tmp_error);
      return;
    }

#if defined(LIBFO_DEBUG) && 0
  fo_object_debug_dump (new_fo_tree, 0);
  fo_object_debug_dump (new_area_tree, 0);
#endif
  *fo_tree = new_fo_tree;
  *area_tree = new_area_tree;

}
